from typing import Any, Optional, Protocol, Self, TypeVar

from .callable_type import CallableCollection, CallableType, StageDefinitionCollection
from .processor_interface import ProcessorInterface
from .traits.clonable import ClonableInterface
from .types import T_in, T_out


class PipelineInterface(
    ClonableInterface,
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        stages: Optional[StageDefinitionCollection] = None,
        stage_instances: Optional[CallableCollection] = None,
        processor: Optional[type[ProcessorInterface] | ProcessorInterface] = None,
        processor_instance: Optional[ProcessorInterface] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    async def process(self, payload: T_in, *args: Any, **kwds: Any) -> T_out: ...

    def pipe(self, stage: CallableType) -> Self: ...

    async def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...

    def with_processor(
        self, processor: type[ProcessorInterface] | ProcessorInterface
    ) -> Self: ...

    def with_stages(self, stages: StageDefinitionCollection) -> Self: ...

    def get_processor_instance(self) -> Optional[ProcessorInterface]: ...

    def get_stages(self) -> StageDefinitionCollection: ...


TPipeline = TypeVar("TPipeline", bound=PipelineInterface, infer_variance=True)



# Version of the package
# DO NOT MODIFY MANUALLY
# This will be updated by `bumpver` command.
# - Make sure to commit all changes first before running `bumpver`.
# - Run `bumpver update --[minor|major|patch]`
__version__ = "1.26.0"

# Re-exporting symbols
from .pipeline import CallableCollection as CallableCollection
from .pipeline import CallableType as CallableType
from .pipeline import Pipeline as Pipeline
from .pipeline import PipelineFactory as PipelineFactory
from .pipeline import PipelineFactoryInterface as PipelineFactoryInterface
from .pipeline import PipelineInterface as PipelineInterface
from .pipeline import Processor as Processor
from .pipeline import ProcessorInterface as ProcessorInterface
from .pipeline import Stage as Stage
from .pipeline import StageDefinition as StageDefinition
from .pipeline import StageDefinitionCollection as StageDefinitionCollection
from .pipeline import StageInterface as StageInterface
from .pipeline import T_in as T_in
from .pipeline import T_out as T_out
from .pipeline import __all__ as _pipeline_all
from .processors import ChainedPipeline as ChainedPipeline
from .processors import ChainedProcessor as ChainedProcessor
from .processors import InterruptiblePipeline as InterruptiblePipeline
from .processors import InterruptibleProcessor as InterruptibleProcessor
from .processors import __all__ as _processor_all

# pyright: reportUnsupportedDunderAll=false
__all__ = (
    *_pipeline_all,
    *_processor_all,
)



from typing import Any, Protocol

from ..callable_type import CallableCollection
from ..processor_interface import ProcessorInterface
from ..types import T_in, T_out


class ChainedProcessorInterface(
    ProcessorInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    """
    A processor that processes the payload through a series of stages.
    """

    async def process(
        self,
        payload: T_in,
        stages: CallableCollection,
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...



from typing import Any, Optional, Self

from .callable_type import CallableCollection, CallableType, StageDefinitionCollection
from .pipeline_interface import PipelineInterface as ImplementsInterface
from .processor_interface import ProcessorInterface
from .processors.chained_processor import ChainedProcessor
from .traits.clonable import Clonable
from .types import T_in, T_out


class Pipeline(
    Clonable,
    ImplementsInterface[T_in, T_out],
):
    stages: StageDefinitionCollection
    stage_instances: CallableCollection
    processor: Optional[type[ProcessorInterface] | ProcessorInterface]
    processor_instance: Optional[ProcessorInterface]

    def __init__(
        self,
        stages: Optional[StageDefinitionCollection] = None,
        stage_instances: Optional[CallableCollection] = None,
        processor: Optional[type[ProcessorInterface] | ProcessorInterface] = None,
        processor_instance: Optional[ProcessorInterface] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        if not hasattr(self, "stages"):
            self.stages = tuple()

        if not hasattr(self, "stage_instances"):
            self.stage_instances = tuple()

        if not hasattr(self, "processor"):
            self.processor = self._get_default_processor()

        if not hasattr(self, "processor_instance"):
            self.processor_instance = None

        if stages:
            self.stages = stages

        if stage_instances:
            self.stage_instances = stage_instances

        if self._should_instantiate_stages():
            self._instantiate_stages()

        if processor:
            self.processor = processor

        if processor_instance:
            self.processor_instance = processor_instance

        if self._should_instantiate_processor():
            self._instantiate_processor()

    async def process(self, payload: T_in, *args: Any, **kwds: Any) -> T_out:
        """
        Process the given payload through the pipeline.
        """
        if self.processor_instance is None:
            raise ValueError("Processor not set")

        return await self.processor_instance.process(
            payload=payload, stages=self.stage_instances, *args, **kwds
        )

    def pipe(self, stage: CallableType) -> Self:
        """
        Adds a single stage to the pipeline.
        """
        return self.clone({"stage_instances": tuple([*self.stage_instances, stage])})

    async def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        """
        Processes payload through the pipeline.
        """
        return await self.process(payload, *args, **kwds)

    def with_processor(
        self, processor: type[ProcessorInterface] | ProcessorInterface
    ) -> Self:
        """
        Attachs a processor (class or instance) to the pipeline.
        """
        cloned = self.clone({"processor": processor, "processor_instance": None})

        return cloned._instantiate_processor()

    def with_stages(self, stages: StageDefinitionCollection) -> Self:
        """
        Adds a collection of stages to the pipeline.
        """
        cloned = self.clone({"stages": stages, "stage_instances": []})

        return cloned._instantiate_stages()

    def get_processor_instance(self) -> Optional[ProcessorInterface]:
        return self.processor_instance

    def get_stages(self) -> StageDefinitionCollection:
        return self.stages

    def _get_default_processor(self) -> type[ChainedProcessor[T_in, T_out]]:
        return ChainedProcessor

    def _should_instantiate_processor(self) -> bool:
        return self.processor_instance is None

    def _instantiate_processor(self) -> Self:
        if self.processor is None:
            raise ValueError("Processor class not set")

        if isinstance(self.processor, type):
            self.processor_instance = self.processor()
        else:
            self.processor_instance = self.processor

        if isinstance(self.processor_instance, type):
            raise ValueError("Processor instance could not be created")

        return self

    def _should_instantiate_stages(self) -> bool:
        return len(self.stage_instances) == 0 and len(self.stages) > 0

    def _instantiate_stages(self) -> Self:
        self.stage_instances = tuple(
            stage() if isinstance(stage, type) else stage for stage in self.stages
        )

        return self



from typing import Any, Optional, Protocol, Self

from .callable_type import StageDefinition, StageDefinitionCollection
from .pipeline_interface import PipelineInterface
from .traits.act_as_factory.act_as_factory_interface import ActAsFactoryInterface
from .types import T_in, T_out


class PipelineFactoryInterface(
    ActAsFactoryInterface[PipelineInterface[T_in, T_out]],
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        stages: Optional[StageDefinitionCollection] = None,
        pipeline_class: Optional[type[PipelineInterface[T_in, T_out]]] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def add_stage(self, stage: StageDefinition) -> Self: ...

    def with_stages(self, stages: StageDefinitionCollection) -> Self: ...



"""Library's public types"""

# Re-exporting symbols
from _lib import CallableCollection as CallableCollection
from _lib import CallableType as CallableType
from _lib import StageDefinition as StageDefinition
from _lib import StageDefinitionCollection as StageDefinitionCollection
from _lib import T_in as T_in
from _lib import T_out as T_out

# pyright: reportUnsupportedDunderAll=false
__all__ = (
    "CallableType",
    "CallableCollection",
    "StageDefinition",
    "StageDefinitionCollection",
    "T_in",
    "T_out",
)
